#!/usr/bin/env node
import chalkTemplate from 'chalk-template';
import util, { parseArgs } from 'util';
import childProcess from 'child_process';
import fs from 'node:fs/promises';
import ora from 'ora';
import path from 'path';
import getPort from 'get-port';
import http from 'http';
import open from 'open';
import toml from '@iarna/toml';
import ini from 'ini';
import os from 'os';

var version = "3.0.6";

const printVersion = ()=>{
    console.log(version);
    process.exit(0);
};
const printHelp = ()=>{
    const text = chalkTemplate`
    {bold.magenta ronin} — Automatically set the types for your RONIN project

    {bold USAGE}

        {bold $} {magenta ronin}
        {bold $} {magenta ronin} login
        {bold $} {magenta ronin} --help
        {bold $} {magenta ronin} --version

    {bold COMMANDS}

        login                               Authenticate with RONIN (run by default for every command)

        init [space]                        Initialize the TypeScript types for a given space

    {bold OPTIONS}

        -h, --help                          Shows this help message

        -v, --version                       Displays the current version of serve

        -d, --debug                         Show debugging information
  `;
    console.log(text);
    process.exit(0);
};

const exec = util.promisify(childProcess.exec);
const exists = async (file)=>{
    try {
        await fs.access(path.join(process.cwd(), file));
    } catch (err) {
        if (err.code === 'ENOENT') {
            return false;
        } else {
            throw err;
        }
    }
    return true;
};
var initializeProject = (async (positionals)=>{
    const spinner = ora('Initializing project').start();
    const lastPositional = positionals[positionals.length - 1];
    const spaceHandle = lastPositional === 'init' ? null : lastPositional;
    if (!spaceHandle) {
        spinner.fail('Please provide a space handle like this:\n$ ronin init my-space');
        process.exit(1);
    }
    if (!await exists('package.json')) {
        spinner.fail('No `package.json` found in the current directory. Please run the command in your project.');
        process.exit(1);
    }
    const packageManager = await exists('bun.lockb') ? 'bun' : 'npm';
    const packageManagerName = packageManager === 'bun' ? 'Bun' : 'npm';
    spinner.text = `Detected ${packageManagerName} — installing types package`;
    try {
        if (packageManager === 'bun') {
            await exec(`bun add @ronin/${spaceHandle} --dev`);
        } else {
            await exec(`npm install @ronin/${spaceHandle} --save-dev`);
        }
    } catch (err) {
        if (err instanceof Error && err.message.includes('401')) {
            spinner.fail(`You are not a member of the "${spaceHandle}" space or the space doesn't exist.`);
            process.exit(1);
        }
    }
    spinner.succeed('Project initialized');
});

const CACHE_DIR = path.join(os.homedir(), '.ronin');
const CACHE_DIR_FILE = path.join(CACHE_DIR, 'session.json');
const getSession = async ()=>{
    try {
        const contents = await fs.readFile(CACHE_DIR_FILE, {
            encoding: 'utf-8'
        });
        const parsedContents = JSON.parse(contents);
        if (parsedContents?.token) return parsedContents;
    } catch (err) {
        if (err.code !== 'ENOENT') throw err;
    }
    return null;
};
const readConfigFile = async (filePath, name, parser)=>{
    let npmConfigContents = {};
    try {
        const contents = await fs.readFile(filePath, {
            encoding: 'utf-8'
        });
        npmConfigContents = parser(contents);
    } catch (err) {
        if (err.code !== 'ENOENT') {
            console.error(`Failed to read ${name} config file at ${filePath}`, err);
        }
    }
    return npmConfigContents;
};
const writeConfigFile = async (filePath, contents)=>{
    return fs.writeFile(filePath, contents, {
        encoding: 'utf-8'
    });
};
const storeSession = async (token)=>{
    // Ensure that the cache directory exists.
    try {
        await fs.stat(CACHE_DIR);
    } catch (err) {
        if (err.code === 'ENOENT') {
            await fs.mkdir(CACHE_DIR);
        } else {
            throw err;
        }
    }
    return writeConfigFile(CACHE_DIR_FILE, JSON.stringify({
        token
    }, null, 2));
};
const storeSessionforNPM = async (token)=>{
    const npmConfigFile = process.env['npm_config_userconfig'] || path.join(os.homedir(), '.npmrc');
    const npmConfigContents = await readConfigFile(npmConfigFile, 'npm', ini.parse);
    npmConfigContents['@ronin:registry'] = 'https://ronin.supply';
    npmConfigContents['//ronin.supply/:_authToken'] = token;
    await writeConfigFile(npmConfigFile, ini.stringify(npmConfigContents));
};
const storeSessionForBun = async (token)=>{
    const bunConfigFile = path.join(os.homedir(), '.bunfig.toml');
    const bunConfigContents = await readConfigFile(bunConfigFile, 'Bun', toml.parse);
    // Safely initialize potentially missing keys.
    if (!bunConfigContents.install) bunConfigContents.install = {};
    if (!bunConfigContents.install.scopes) bunConfigContents.install.scopes = {};
    if (!bunConfigContents.install.scopes.ronin) bunConfigContents.install.scopes.ronin = {};
    bunConfigContents.install.scopes.ronin.url = 'https://ronin.supply';
    bunConfigContents.install.scopes.ronin.token = token;
    await writeConfigFile(bunConfigFile, toml.stringify(bunConfigContents));
};

var logIn = (async ()=>{
    const spinner = ora('Logging in').start();
    let token;
    const port = await getPort();
    const server = http.createServer().listen(port);
    const baseURL = new URL('https://ronin.co/actions/auth/cli');
    const currentHost = `http://localhost:${port}`;
    const initURL = new URL(baseURL);
    initURL.searchParams.set('from', currentHost);
    spinner.text = `Please log in using the following URL:\n${initURL.href}`;
    spinner.suffixText = `\n`;
    try {
        [token] = await Promise.all([
            new Promise((resolve, reject)=>{
                server.once('request', (req, res)=>{
                    res.setHeader('connection', 'close');
                    const requestURL = new URL(req.url || '', currentHost);
                    const requestToken = requestURL.searchParams.get('token');
                    if (!requestToken) return reject(new Error('Missing token'));
                    resolve(requestToken);
                    const headers = {
                        Location: baseURL.href
                    };
                    res.writeHead(307, headers).end();
                });
                server.once('error', reject);
            }),
            open(initURL.href)
        ]);
    } finally{
        server.close();
    }
    // Initialize sessions for the RONIN CLI and all relevant package managers.
    await Promise.all([
        storeSession(token),
        storeSessionforNPM(token),
        storeSessionForBun(token)
    ]);
    spinner.suffixText = '';
    spinner.succeed('Logged in successfully!');
});

let values;
let positionals;
try {
    ({ values, positionals } = parseArgs({
        args: process.argv,
        options: {
            help: {
                type: 'boolean',
                short: 'h',
                default: false
            },
            version: {
                type: 'boolean',
                short: 'v',
                default: false
            },
            debug: {
                type: 'boolean',
                short: 'd',
                default: false
            }
        },
        strict: true,
        allowPositionals: true
    }));
} catch (err) {
    if (err instanceof Error) {
        console.error(err.message);
    } else {
        throw err;
    }
    process.exit(1);
}
const run = async ()=>{
    // Flags for printing useful information about the CLI.
    if (values.help) return printHelp();
    if (values.version) return printVersion();
    // If there is no active session, automatically start one and then continue
    // with the execution of the requested sub command, if there is one.
    const session = await getSession();
    if (!session) await logIn();
    // This ensures that people can accidentally type uppercase letters and still
    // get the command they are looking for.
    const normalizedPositionals = positionals.map((positional)=>positional.toLowerCase());
    // `login` command
    if (normalizedPositionals.includes('login')) return logIn();
    // `init` command
    if (normalizedPositionals.includes('init')) return initializeProject(positionals);
    // If no matching flags or commands were found, render the help, since we
    // don't want to use the main `ronin` command for anything yet.
    return printHelp();
};
run();
